@@ -1,4 +1,4 @@
-/* jdupes (C) 2015-2022 Jody Bruchon <jody@jodybruchon.com>
+/* jdupes (C) 2015-2023 Jody Bruchon <jody@jodybruchon.com>
    Forked from fdupes 1.51 (C) 1999-2014 Adrian Lopez
 
    Permission is hereby granted, free of charge, to any person
@@ -42,7 +42,11 @@
 #include <time.h>
 #include <sys/time.h>
 #include "jdupes.h"
-#include "xxhash.h"
+#ifndef USE_JODY_HASH
+ #include "xxhash.h"
+#else
+ #include "jody_hash.h"
+#endif
 #include "oom.h"
 #ifdef ENABLE_DEDUPE
 #include <sys/utsname.h>
@@ -61,7 +65,9 @@
 #include "act_dedupefiles.h"
 #include "act_linkfiles.h"
 #include "act_printmatches.h"
-#include "act_printjson.h"
+#ifndef NO_JSON
+ #include "act_printjson.h"
+#endif /* NO_JSON */
 #include "act_summarize.h"
 
 /* Detect Windows and modify as needed */
@@ -106,17 +112,22 @@ static const char *program_name;
  static int usr1_toggle = 0;
 #endif
 
-/* Larger chunk size makes large files process faster but uses more RAM */
-#define MIN_CHUNK_SIZE 4096
-#define MAX_CHUNK_SIZE 16777216
-#ifndef CHUNK_SIZE
- #define CHUNK_SIZE 65536
-#endif
 #ifndef PARTIAL_HASH_SIZE
  #define PARTIAL_HASH_SIZE 4096
 #endif
+#ifndef CHUNK_SIZE
+ #define CHUNK_SIZE 65536
+#endif
 
-static size_t auto_chunk_size = CHUNK_SIZE;
+#ifndef NO_CHUNKSIZE
+ /* Larger chunk size makes large files process faster but uses more RAM */
+ #define MIN_CHUNK_SIZE 4096
+ #define MAX_CHUNK_SIZE 16777216
+ static size_t auto_chunk_size = CHUNK_SIZE;
+#else
+ /* If automatic chunk sizing is disabled, just use a fixed value */
+ #define auto_chunk_size CHUNK_SIZE
+#endif /* NO_CHUNKSIZE */
 
 /* Maximum path buffer size to use; must be large enough for a path plus
  * any work that might be done to the array it's stored in. PATH_MAX is
@@ -202,15 +213,21 @@ const char *extensions[] = {
     NULL
 };
 
+#ifndef NO_TRAVCHECK
+/* Simple traversal balancing hash - scrambles inode number */
+#define TRAVHASH(device,inode) (((inode << 55 | (inode >> 9)) + (device << 13)))
 /* Tree to track each directory traversed */
 struct travdone {
   struct travdone *left;
   struct travdone *right;
+  uintmax_t hash;
   jdupes_ino_t inode;
   dev_t device;
 };
 static struct travdone *travdone_head = NULL;
+#endif /* NO_TRAVCHECK */
 
+#ifndef NO_EXTFILTER
 /* Extended filter tree head and static tag list */
 struct extfilter *extfilter_head = NULL;
 const struct extfilter_tags extfilter_tags[] = {
@@ -227,6 +244,8 @@ const struct extfilter_tags extfilter_tags[] = {
   { "older",	XF_DATE_OLDER },
   { NULL, 0 },
 };
+static void help_text_extfilter(void);
+#endif /* NO_EXTFILTER */
 
 /* Required for progress indicator code */
 static uintmax_t filecount = 0;
@@ -234,22 +253,25 @@ static uintmax_t progress = 0, item_progress = 0, dupecount = 0;
 /* Number of read loops before checking progress indicator */
 #define CHECK_MINIMUM 256
 
-/* Hash/compare performance statistics (debug mode) */
+/* Performance and behavioral statistics (debug mode) */
 #ifdef DEBUG
 static unsigned int small_file = 0, partial_hash = 0, partial_elim = 0;
 static unsigned int full_hash = 0, partial_to_full = 0, hash_fail = 0;
 static uintmax_t comparisons = 0;
-static unsigned int left_branch = 0, right_branch = 0;
+static unsigned int filetree_lefts = 0, filetree_rights = 0;
  #ifdef ON_WINDOWS
   #ifndef NO_HARDLINKS
-static unsigned int hll_exclude = 0;
+  static unsigned int hll_exclude = 0;
   #endif
  #endif
+ #ifndef NO_TRAVCHECK
+ static uintmax_t travdone_lefts = 0, travdone_rights = 0;
+ #endif
 #endif /* DEBUG */
 
 #ifdef TREE_DEPTH_STATS
-static unsigned int tree_depth = 0;
-static unsigned int max_depth = 0;
+static unsigned int filetree_depth = 0;
+static unsigned int filetree_max_depth = 0;
 #endif
 
 /* File tree head */
@@ -276,8 +298,6 @@ char tempname[PATHBUF_SIZE * 2];
 /* Compare two hashes like memcmp() */
 #define HASH_COMPARE(a,b) ((a > b) ? 1:((a == b) ? 0:-1))
 
-static void help_text_extfilter(void);
-
 /***** End definitions, begin code *****/
 
 /* Catch CTRL-C and either notify or terminate */
@@ -419,6 +439,7 @@ static void update_progress(const char * const restrict msg, const int file_perc
 /***** Add new functions here *****/
 
 
+#ifndef NO_EXTFILTER
 /* Does a file have one of these comma-separated extensions?
  * Returns 1 after any match, 0 if no matches */
 int match_extensions(char *path, const char *extlist)
@@ -445,7 +466,7 @@ int match_extensions(char *path, const char *extlist)
 
   /* Get the length of the file's extension for later checking */
   extlen = strlen(dot);
-  LOUD(fprintf(stderr, "match_extensions: file has extension '%s' with length %ld\n", dot, extlen);)
+  LOUD(fprintf(stderr, "match_extensions: file has extension '%s' with length %" PRIdMAX "\n", dot, (intmax_t)extlen);)
 
   /* dot is now at the location of the last file extension; check the list */
   /* Skip any commas at the start of the list */
@@ -461,10 +482,10 @@ int match_extensions(char *path, const char *extlist)
       while (*extlist == ',') extlist++;
       if (extlist == ext)  goto skip_empty;
       if (strncasecmp(dot, ext, len) == 0 && extlen == len) {
-        LOUD(fprintf(stderr, "match_extensions: matched on extension '%s' (len %ld)\n", dot, len);)
+        LOUD(fprintf(stderr, "match_extensions: matched on extension '%s' (len %" PRIdMAX ")\n", dot, (intmax_t)len);)
         return 1;
       }
-      LOUD(fprintf(stderr, "match_extensions: no match: '%s' (%ld), '%s' (%ld)\n", dot, len, ext, extlen);)
+      LOUD(fprintf(stderr, "match_extensions: no match: '%s' (%" PRIdMAX "), '%s' (%" PRIdMAX ")\n", dot, (intmax_t)len, ext, (intmax_t)extlen);)
 skip_empty:
       ext = extlist;
       len = 0;
@@ -475,6 +496,7 @@ skip_empty:
   }
   return 0;
 }
+#endif /* NO_EXTFILTER */
 
 
 /* Check file's stat() info to make sure nothing has changed
@@ -544,7 +566,7 @@ int getfilestats(file_t * const restrict file)
   return 0;
 }
 
-
+#ifndef NO_EXTFILTER
 static void add_extfilter(const char *option)
 {
   char *opt, *p;
@@ -612,9 +634,10 @@ static void add_extfilter(const char *option)
       extf->size *= ss->multiplier;
     }
   } else if (extf->flags & XF_REQ_DATE) {
+    /* Exclude uses a date; convert it to seconds since the epoch */
     *(extf->param) = '\0';
     tt = strtoepoch(p);
-    LOUD(fprintf(stderr, "extfilter: jody_strtoepoch: '%s' -> %ld\n", p, tt);)
+    LOUD(fprintf(stderr, "extfilter: jody_strtoepoch: '%s' -> %" PRIdMAX "\n", p, (intmax_t)tt);)
     if (tt == -1) goto error_bad_time;
     extf->size = tt;
   } else {
@@ -645,6 +668,7 @@ error_bad_size_suffix:
   help_text_extfilter();
   exit(EXIT_FAILURE);
 }
+#endif /* NO_EXTFILTER */
 
 
 /* Returns -1 if stat() fails, 0 if it's a directory, 1 if it's not */
@@ -740,7 +764,9 @@ extern int check_conditions(const file_t * const restrict file1, const file_t *
 static int check_singlefile(file_t * const restrict newfile)
 {
   char * restrict tp = tempname;
+#ifndef NO_EXTFILTER
   int excluded;
+#endif /* NO_EXTFILTER */
 
   if (newfile == NULL) nullptr("check_singlefile()");
 
@@ -771,11 +797,12 @@ static int check_singlefile(file_t * const restrict newfile)
     return 1;
   }
 
-    /* Exclude files based on exclusion stack size specs */
+#ifndef NO_EXTFILTER
+    /* Exclude files based on extended filter stack */
     excluded = 0;
     for (struct extfilter *extf = extfilter_head; extf != NULL; extf = extf->next) {
       uint32_t sflag = extf->flags;
-      LOUD(fprintf(stderr, "check_singlefile: extfilter check: %08x %ld %ld %s\n", sflag, newfile->size, extf->size, newfile->d_name);)
+      LOUD(fprintf(stderr, "check_singlefile: extfilter check: %08x %" PRIdMAX " %" PRIdMAX " %s\n", sflag, (intmax_t)newfile->size, (intmax_t)extf->size, newfile->d_name);)
       if (
            /* Any line that passes will result in file exclusion */
            ((sflag == XF_SIZE_EQ) && (newfile->size != extf->size))
@@ -797,6 +824,7 @@ static int check_singlefile(file_t * const restrict newfile)
       LOUD(fprintf(stderr, "check_singlefile: excluding based on an extfilter option\n"));
       return 1;
     }
+#endif /* NO_EXTFILTER */
   }
 
 #ifdef ON_WINDOWS
@@ -824,7 +852,7 @@ static file_t *init_newfile(const size_t len, file_t * restrict * const restrict
   if (!newfile) oom("init_newfile() file structure");
   if (!filelistp) nullptr("init_newfile() filelistp");
 
-  LOUD(fprintf(stderr, "init_newfile(len %lu, filelistp %p)\n", len, filelistp));
+  LOUD(fprintf(stderr, "init_newfile(len %" PRIuMAX", filelistp %p)\n", (uintmax_t)len, filelistp));
 
   memset(newfile, 0, sizeof(file_t));
   newfile->d_name = (char *)string_malloc(len);
@@ -840,12 +868,13 @@ static file_t *init_newfile(const size_t len, file_t * restrict * const restrict
 }
 
 
+#ifndef NO_TRAVCHECK
 /* Create a new traversal check object and initialize its values */
-static struct travdone *travdone_alloc(const dev_t device, const jdupes_ino_t inode)
+static struct travdone *travdone_alloc(const dev_t device, const jdupes_ino_t inode, uintmax_t hash)
 {
   struct travdone *trav;
 
-  LOUD(fprintf(stderr, "travdone_alloc(%" PRIdMAX ", %" PRIdMAX ")\n", (intmax_t)inode, (intmax_t)device);)
+  LOUD(fprintf(stderr, "travdone_alloc(dev %" PRIdMAX ", ino %" PRIdMAX ", hash %" PRIuMAX ")\n", (intmax_t)device, (intmax_t)inode, hash);)
 
   trav = (struct travdone *)string_malloc(sizeof(struct travdone));
   if (trav == NULL) {
@@ -854,8 +883,9 @@ static struct travdone *travdone_alloc(const dev_t device, const jdupes_ino_t in
   }
   trav->left = NULL;
   trav->right = NULL;
-  trav->inode = inode;
+  trav->hash = hash;
   trav->device = device;
+  trav->inode = inode;
   LOUD(fprintf(stderr, "travdone_alloc returned %p\n", (void *)trav);)
   return trav;
 }
@@ -865,11 +895,22 @@ static struct travdone *travdone_alloc(const dev_t device, const jdupes_ino_t in
 static void travdone_free(struct travdone * const restrict cur)
 {
   LOUD(fprintf(stderr, "travdone_free(%p)\n", cur);)
-  if (cur == NULL) return;
+  if (cur == NULL) goto error_travdone_null;
+  if (cur->left == cur) goto error_travdone_left;
+  if (cur->right == cur) goto error_travdone_right;
   if (cur->left != NULL) travdone_free(cur->left);
   if (cur->right != NULL) travdone_free(cur->right);
   string_free(cur);
   return;
+error_travdone_null:
+  fprintf(stderr, "internal error: travdone_free() was passed NULL which should never happen\n");
+  exit(EXIT_FAILURE);
+error_travdone_left:
+  fprintf(stderr, "internal error: travdone_free() was passed NULL which should never happen\n");
+  exit(EXIT_FAILURE);
+error_travdone_right:
+  fprintf(stderr, "internal error: travdone_free() was passed NULL which should never happen\n");
+  exit(EXIT_FAILURE);
 }
 
 
@@ -877,9 +918,12 @@ static void travdone_free(struct travdone * const restrict cur)
 static int traverse_check(const dev_t device, const jdupes_ino_t inode)
 {
   struct travdone *traverse = travdone_head;
+  uintmax_t travhash;
 
+  LOUD(fprintf(stderr, "traverse_check(dev %" PRIuMAX ", ino %" PRIuMAX "\n", (uintmax_t)device, (uintmax_t)inode);)
+  travhash = TRAVHASH(device, inode);
   if (travdone_head == NULL) {
-    travdone_head = travdone_alloc(device, inode);
+    travdone_head = travdone_alloc(device, inode, TRAVHASH(device, inode));
     if (travdone_head == NULL) return 2;
   } else {
     traverse = travdone_head;
@@ -887,33 +931,38 @@ static int traverse_check(const dev_t device, const jdupes_ino_t inode)
       if (traverse == NULL) nullptr("traverse_check()");
       /* Don't re-traverse directories we've already seen */
       if (inode == traverse->inode && device == traverse->device) {
-        LOUD(fprintf(stderr, "traverse_check: already seen: %ld:%ld\n", device,inode);)
+        LOUD(fprintf(stderr, "traverse_check: already seen: %" PRIuMAX ":%" PRIuMAX "\n", (uintmax_t)device, (uintmax_t)inode);)
         return 1;
-      } else if (inode > traverse->inode || (inode == traverse->inode && device > traverse->device)) {
-        /* Traverse right */
-        if (traverse->right == NULL) {
-          LOUD(fprintf(stderr, "traverse item right: %ld:%ld\n", device, inode);)
-          traverse->right = travdone_alloc(device, inode);
-          if (traverse->right == NULL) return 2;
-          break;
-        }
-        traverse = traverse->right;
-        continue;
       } else {
-        /* Traverse left */
-        if (traverse->left == NULL) {
-          LOUD(fprintf(stderr, "traverse item left %ld,%ld\n", device, inode);)
-          traverse->left = travdone_alloc(device, inode);
-          if (traverse->left == NULL) return 2;
-          break;
+        if (travhash > traverse->hash) {
+          /* Traverse right */
+          if (traverse->right == NULL) {
+            LOUD(fprintf(stderr, "traverse_check add right: %" PRIuMAX ", %" PRIuMAX"\n", (uintmax_t)device, (uintmax_t)inode);)
+            DBG(travdone_rights++);
+            traverse->right = travdone_alloc(device, inode, travhash);
+            if (traverse->right == NULL) return 2;
+            break;
+          }
+          traverse = traverse->right;
+          continue;
+        } else {
+          /* Traverse left */
+          if (traverse->left == NULL) {
+            LOUD(fprintf(stderr, "traverse_check add left: %" PRIuMAX ", %" PRIuMAX "\n", (uintmax_t)device, (uintmax_t)inode);)
+            DBG(travdone_lefts++);
+            traverse->left = travdone_alloc(device, inode, travhash);
+            if (traverse->left == NULL) return 2;
+            break;
+          }
+          traverse = traverse->left;
+          continue;
         }
-        traverse = traverse->left;
-        continue;
       }
     }
   }
   return 0;
 }
+#endif /* NO_TRAVCHECK */
 
 
 /* This is disabled until a check is in place to make it safe */
@@ -991,12 +1040,14 @@ static void grokdir(const char * const restrict dir,
     return; /* Remove when single file is restored */
   }
 
+#ifndef NO_TRAVCHECK
   /* Double traversal prevention tree */
   if (!ISFLAG(flags, F_NOTRAVCHECK)) {
     i = traverse_check(device, inode);
     if (i == 1) return;
     if (i == 2) goto error_travdone;
   }
+#endif /* NO_TRAVCHECK */
 
   item_progress++;
   grokdir_level++;
@@ -1167,7 +1218,10 @@ static jdupes_hash_t *get_filehash(const file_t * const restrict checkfile,
   static jdupes_hash_t *chunk = NULL;
   FILE *file;
   int check = 0;
+#ifndef USE_JODY_HASH
   XXH64_state_t *xxhstate;
+#else
+#endif
 
   if (checkfile == NULL || checkfile->d_name == NULL) nullptr("get_filehash()");
   LOUD(fprintf(stderr, "get_filehash('%s', %" PRIdMAX ")\n", checkfile->d_name, (intmax_t)max_read);)
@@ -1227,11 +1281,13 @@ static jdupes_hash_t *get_filehash(const file_t * const restrict checkfile,
     fsize -= PARTIAL_HASH_SIZE;
   }
 
+#ifndef USE_JODY_HASH
   xxhstate = XXH64_createState();
   if (xxhstate == NULL) nullptr("xxhstate");
   XXH64_reset(xxhstate, 0);
+#endif
 
-  /* Read the file in CHUNK_SIZE chunks until we've read it all. */
+  /* Read the file in chunks until we've read it all. */
   while (fsize > 0) {
     size_t bytes_to_read;
 
@@ -1243,7 +1299,11 @@ static jdupes_hash_t *get_filehash(const file_t * const restrict checkfile,
       return NULL;
     }
 
+#ifndef USE_JODY_HASH
     XXH64_update(xxhstate, chunk, bytes_to_read);
+#else
+    *hash = jody_block_hash(chunk, *hash, bytes_to_read);
+#endif
 
     if ((off_t)bytes_to_read > fsize) break;
     else fsize -= (off_t)bytes_to_read;
@@ -1259,8 +1319,10 @@ static jdupes_hash_t *get_filehash(const file_t * const restrict checkfile,
 
   fclose(file);
 
+#ifndef USE_JODY_HASH
   *hash = XXH64_digest(xxhstate);
   XXH64_freeState(xxhstate);
+#endif
 
   LOUD(fprintf(stderr, "get_filehash: returning hash: 0x%016jx\n", (uintmax_t)*hash));
   return hash;
@@ -1307,7 +1369,7 @@ static inline void registerfile(filetree_t * restrict * const restrict nodeptr,
 
 
 #ifdef TREE_DEPTH_STATS
-#define TREE_DEPTH_UPDATE_MAX() { if (max_depth < tree_depth) max_depth = tree_depth; tree_depth = 0; }
+#define TREE_DEPTH_UPDATE_MAX() { if (filetree_max_depth < filetree_depth) filetree_max_depth = filetree_depth; filetree_depth = 0; }
 #else
 #define TREE_DEPTH_UPDATE_MAX()
 #endif
@@ -1448,7 +1510,7 @@ static file_t **checkmatch(filetree_t * restrict tree, file_t * const restrict f
   if (cmpresult < 0) {
     if (tree->left != NULL) {
       LOUD(fprintf(stderr, "checkmatch: recursing tree: left\n"));
-      DBG(left_branch++; tree_depth++;)
+      DBG(filetree_lefts++; filetree_depth++;)
       return checkmatch(tree->left, file);
     } else {
       LOUD(fprintf(stderr, "checkmatch: registering file: left\n"));
@@ -1459,7 +1521,7 @@ static file_t **checkmatch(filetree_t * restrict tree, file_t * const restrict f
   } else if (cmpresult > 0) {
     if (tree->right != NULL) {
       LOUD(fprintf(stderr, "checkmatch: recursing tree: right\n"));
-      DBG(right_branch++; tree_depth++;)
+      DBG(filetree_rights++; filetree_depth++;)
       return checkmatch(tree->right, file);
     } else {
       LOUD(fprintf(stderr, "checkmatch: registering file: right\n"));
@@ -1578,8 +1640,12 @@ static int sort_pairs_by_mtime(file_t *f1, file_t *f2)
   if (f1->mtime < f2->mtime) return -sort_direction;
   else if (f1->mtime > f2->mtime) return sort_direction;
 
+#ifndef NO_NUMSORT
   /* If the mtimes match, use the names to break the tie */
   return numeric_sort(f1->d_name, f2->d_name, sort_direction);
+#else
+  return strcmp(f1->d_name, f2->d_name) ? -sort_direction : sort_direction;
+#endif /* NO_NUMSORT */
 }
 #endif
 
@@ -1593,7 +1659,11 @@ static int sort_pairs_by_filename(file_t *f1, file_t *f2)
   if (po != 0) return po;
 #endif /* NO_USER_ORDER */
 
+#ifndef NO_NUMSORT
   return numeric_sort(f1->d_name, f2->d_name, sort_direction);
+#else
+  return strcmp(f1->d_name, f2->d_name) ? -sort_direction : sort_direction;
+#endif /* NO_NUMSORT */
 }
 
 
@@ -1657,13 +1727,17 @@ static inline void help_text(void)
 #ifdef ENABLE_DEDUPE
   printf(" -B --dedupe      \tdo a copy-on-write (reflink/clone) deduplication\n");
 #endif
+#ifndef NO_CHUNKSIZE
   printf(" -C --chunk-size=#\toverride I/O chunk size (min %d, max %d)\n", MIN_CHUNK_SIZE, MAX_CHUNK_SIZE);
+#endif /* NO_CHUNKSIZE */
+#ifndef NO_DELETE
   printf(" -d --delete      \tprompt user for files to preserve and delete all\n");
   printf("                  \tothers; important: under particular circumstances,\n");
   printf("                  \tdata may be lost when using this option together\n");
   printf("                  \twith -s or --symlinks, or when specifying a\n");
   printf("                  \tparticular directory more than once; refer to the\n");
   printf("                  \tdocumentation for additional information\n");
+#endif /* NO_DELETE */
 #ifdef DEBUG
   printf(" -D --debug       \toutput debug statistics after completion\n");
 #endif
@@ -1677,7 +1751,9 @@ static inline void help_text(void)
 #ifndef NO_USER_ORDER
   printf(" -I --isolate     \tfiles in the same specified directory won't match\n");
 #endif
+#ifndef NO_JSON
   printf(" -j --json        \tproduce JSON (machine-readable) output\n");
+#endif /* NO_JSON */
 /*  printf(" -K --skip-hash   \tskip full file hashing (may be faster; 100%% safe)\n");
     printf("                  \tWARNING: in development, not fully working yet!\n"); */
 #ifndef NO_SYMLINKS
@@ -1693,9 +1769,11 @@ static inline void help_text(void)
 #endif /* NO_HARDLINKS */
   printf(" -m --summarize   \tsummarize dupe information\n");
   printf(" -M --print-summarize\tprint match sets and --summarize at the end\n");
+#ifndef NO_DELETE
   printf(" -N --no-prompt   \ttogether with --delete, preserve the first file in\n");
   printf("                  \teach set of duplicates and delete the rest without\n");
   printf("                  \tprompting the user\n");
+#endif /* NO_DELETE */
 #ifndef NO_MTIME
   printf(" -o --order=BY    \tselect sort order for output, linking and deleting; by\n");
   printf("                  \tmtime (BY=time) or filename (BY=name, the default)\n");
@@ -1727,8 +1805,10 @@ static inline void help_text(void)
   printf(" -U --no-trav-check\tdisable double-traversal safety check (BE VERY CAREFUL)\n");
   printf("                  \tThis fixes a Google Drive File Stream recursion issue\n");
   printf(" -v --version     \tdisplay jdupes version and license information\n");
+#ifndef NO_EXTFILTER
   printf(" -X --ext-filter=x:y\tfilter files based on specified criteria\n");
   printf("                  \tUse '-X help' for detailed extfilter help\n");
+#endif /* NO_EXTFILTER */
   printf(" -z --zero-match  \tconsider zero-length files to be duplicates\n");
   printf(" -Z --soft-abort  \tIf the user aborts (i.e. CTRL-C) act on matches so far\n");
 #ifndef ON_WINDOWS
@@ -1740,6 +1820,7 @@ static inline void help_text(void)
 }
 
 
+#ifndef NO_EXTFILTER
 static void help_text_extfilter(void)
 {
   printf("Detailed help for jdupes -X/--ext-filter options\n");
@@ -1773,6 +1854,7 @@ static void help_text_extfilter(void)
   printf(  "Extension matching is case-insensitive.\n");
   printf(  "Path substring matching is case-sensitive.\n");
 }
+#endif /* NO_EXTFILTER */
 
 
 #ifdef UNICODE
@@ -1791,7 +1873,9 @@ int main(int argc, char **argv)
 #ifndef NO_MTIME  /* Remove if new order types are added! */
   static ordertype_t ordertype = ORDER_NAME;
 #endif
+#ifndef NO_CHUNKSIZE
   static long manual_chunk_size = 0;
+#endif /* NO_CHUNKSIZE */
 #ifdef __linux__
   static struct proc_cacheinfo pci;
 #endif
@@ -1806,6 +1890,7 @@ int main(int argc, char **argv)
     { "printnull", 0, 0, '0' }, //LEGACY
     { "print-null", 0, 0, '0' },
     { "one-file-system", 0, 0, '1' },
+    { "benchmark-stop", 0, 0, '9' },
     { "nohidden", 0, 0, 'A' }, //LEGACY
     { "no-hidden", 0, 0, 'A' },
     { "dedupe", 0, 0, 'B' },
@@ -1863,7 +1948,7 @@ int main(int argc, char **argv)
 #define GETOPT getopt
 #endif
 
-#define GETOPT_STRING "@01ABC:DdfHhIijKLlMmNnOo:P:pQqRrSsTtUuVvX:Zz"
+#define GETOPT_STRING "@019ABC:DdfHhIijKLlMmNnOo:P:pQqRrSsTtUuVvX:Zz"
 
 /* Windows buffers our stderr output; don't let it do that */
 #ifdef ON_WINDOWS
@@ -1886,6 +1971,7 @@ int main(int argc, char **argv)
   else err_mode = _O_U16TEXT;
 #endif /* UNICODE */
 
+#ifndef NO_CHUNKSIZE
 #ifdef __linux__
   /* Auto-tune chunk size to be half of L1 data cache if possible */
   get_proc_cacheinfo(&pci);
@@ -1897,6 +1983,7 @@ int main(int argc, char **argv)
   if ((auto_chunk_size & 0x00000fffUL) != 0)
     auto_chunk_size = (auto_chunk_size + 0x00000fffUL) & 0x000ff000;
 #endif /* __linux__ */
+#endif /* NO_CHUNKSIZE */
 
   /* Is stderr a terminal? If not, we won't write progress to it */
 #ifdef ON_WINDOWS
@@ -1925,9 +2012,13 @@ int main(int argc, char **argv)
       SETFLAG(flags, F_ONEFS);
       LOUD(fprintf(stderr, "opt: recursion across filesystems disabled (--one-file-system)\n");)
       break;
+    case '9':
+      SETFLAG(flags, F_BENCHMARKSTOP);
+      break;
     case 'A':
       SETFLAG(flags, F_EXCLUDEHIDDEN);
       break;
+#ifndef NO_CHUNKSIZE
     case 'C':
       manual_chunk_size = strtol(optarg, NULL, 10) & 0x0ffff000L;  /* Align to 4K sizes */
       if (manual_chunk_size < MIN_CHUNK_SIZE || manual_chunk_size > MAX_CHUNK_SIZE) {
@@ -1937,10 +2028,13 @@ int main(int argc, char **argv)
       } else auto_chunk_size = (size_t)manual_chunk_size;
       LOUD(fprintf(stderr, "Manual chunk size is %ld\n", manual_chunk_size));
       break;
+#endif /* NO_CHUNKSIZE */
+#ifndef NO_DELETE
     case 'd':
       SETFLAG(a_flags, FA_DELETEFILES);
       LOUD(fprintf(stderr, "opt: delete files after matching (--delete)\n");)
       break;
+#endif /* NO_DELETE */
     case 'D':
 #ifdef DEBUG
       SETFLAG(flags, F_DEBUG);
@@ -1983,10 +2077,12 @@ int main(int argc, char **argv)
       fprintf(stderr, "warning: -I and -O are disabled and ignored in this build\n");
       break;
 #endif
+#ifndef NO_JSON
     case 'j':
       SETFLAG(a_flags, FA_PRINTJSON);
       LOUD(fprintf(stderr, "opt: print output in JSON format (--print-json)\n");)
       break;
+#endif /* NO_JSON */
     case 'K':
       SETFLAG(flags, F_SKIPHASH);
       break;
@@ -1999,10 +2095,12 @@ int main(int argc, char **argv)
       SETFLAG(a_flags, FA_PRINTMATCHES);
       LOUD(fprintf(stderr, "opt: print matches with a summary (--print-summarize)\n");)
       break;
+#ifndef NO_DELETE
     case 'N':
       SETFLAG(flags, F_NOPROMPT);
       LOUD(fprintf(stderr, "opt: delete files without prompting (--noprompt)\n");)
       break;
+#endif /* NO_DELETE */
     case 'p':
       SETFLAG(flags, F_PERMISSIONS);
       LOUD(fprintf(stderr, "opt: permissions must also match (--permissions)\n");)
@@ -2068,9 +2166,11 @@ int main(int argc, char **argv)
       SETFLAG(a_flags, FA_SHOWSIZE);
       LOUD(fprintf(stderr, "opt: show size of files enabled (--size)\n");)
       break;
+#ifndef NO_EXTFILTER
     case 'X':
       add_extfilter(optarg);
       break;
+#endif /* NO_EXTFILTER */
     case 'z':
       SETFLAG(flags, F_INCLUDEEMPTY);
       LOUD(fprintf(stderr, "opt: zero-length files count as matches (--zero-match)\n");)
@@ -2107,7 +2207,7 @@ int main(int argc, char **argv)
           c++;
         }
       } else printf(" none");
-      printf("\nCopyright (C) 2015-2022 by Jody Bruchon and contributors\n");
+      printf("\nCopyright (C) 2015-2023 by Jody Bruchon and contributors\n");
       printf("Forked from fdupes 1.51, (C) 1999-2014 Adrian Lopez and contributors\n\n");
       printf("Permission is hereby granted, free of charge, to any person obtaining a copy of\n");
       printf("this software and associated documentation files (the \"Software\"), to deal in\n");
@@ -2264,8 +2364,16 @@ int main(int argc, char **argv)
     }
   }
 
+#ifndef NO_TRAVCHECK
   /* We don't need the double traversal check tree anymore */
   travdone_free(travdone_head);
+#endif /* NO_TRAVCHECK */
+
+  /* Pass -9 option to exit after traversal/loading code */
+  if (ISFLAG(flags, F_BENCHMARKSTOP)) {
+    fprintf(stderr, "\nBenchmarking stop requested; exiting.\n");
+    goto skip_all_scan_code;
+  }
 
   if (ISFLAG(flags, F_REVERSESORT)) sort_direction = -1;
   if (!ISFLAG(flags, F_HIDEPROGRESS)) fprintf(stderr, "\n");
@@ -2369,13 +2477,15 @@ skip_full_check:
 skip_file_scan:
   /* Stop catching CTRL+C */
   signal(SIGINT, SIG_DFL);
+#ifndef NO_DELETE
   if (ISFLAG(a_flags, FA_DELETEFILES)) {
     if (ISFLAG(flags, F_NOPROMPT)) deletefiles(files, 0, 0);
     else deletefiles(files, 1, stdin);
   }
+#endif /* NO_DELETE */
 #ifndef NO_SYMLINKS
   if (ISFLAG(a_flags, FA_MAKESYMLINKS)) linkfiles(files, 0, 0);
-#endif
+#endif /* NO_SYMLINKS */
 #ifndef NO_HARDLINKS
   if (ISFLAG(a_flags, FA_HARDLINKFILES)) linkfiles(files, 1, 0);
 #endif /* NO_HARDLINKS */
@@ -2384,12 +2494,15 @@ skip_file_scan:
 #endif /* ENABLE_DEDUPE */
   if (ISFLAG(a_flags, FA_PRINTMATCHES)) printmatches(files);
   if (ISFLAG(a_flags, FA_PRINTUNIQUE)) printunique(files);
+#ifndef NO_JSON
   if (ISFLAG(a_flags, FA_PRINTJSON)) printjson(files, argc, argv);
+#endif /* NO_JSON */
   if (ISFLAG(a_flags, FA_SUMMARIZEMATCHES)) {
     if (ISFLAG(a_flags, FA_PRINTMATCHES)) printf("\n\n");
     summarizematches(files);
   }
 
+skip_all_scan_code:
   string_malloc_destroy();
 
 #ifdef DEBUG
@@ -2398,12 +2511,16 @@ skip_file_scan:
         partial_hash, small_file, full_hash, partial_to_full,
         partial_elim, hash_fail, (unsigned int)sizeof(jdupes_hash_t)*8);
     fprintf(stderr, "%" PRIuMAX " total files, %" PRIuMAX " comparisons, branch L %u, R %u, both %u, max tree depth %u\n",
-        filecount, comparisons, left_branch, right_branch,
-        left_branch + right_branch, max_depth);
+        filecount, comparisons, filetree_lefts, filetree_rights,
+        filetree_lefts + filetree_rights, filetree_max_depth);
+#ifndef NO_TRAVCHECK
+    fprintf(stderr, "travdone allocs: left %" PRIuMAX ", right %" PRIuMAX " = heaviness %" PRIdMAX "\n", travdone_lefts, travdone_rights, (uintmax_t)(travdone_rights - travdone_lefts));
+#endif
     fprintf(stderr, "SMA: allocs %" PRIuMAX ", free %" PRIuMAX " (merge %" PRIuMAX ", repl %" PRIuMAX "), fail %" PRIuMAX ", reuse %" PRIuMAX ", scan %" PRIuMAX ", tails %" PRIuMAX "\n",
         sma_allocs, sma_free_good, sma_free_merged, sma_free_replaced,
         sma_free_ignored, sma_free_reclaimed,
         sma_free_scanned, sma_free_tails);
+#ifndef NO_CHUNKSIZE
     if (manual_chunk_size > 0) fprintf(stderr, "I/O chunk size: %ld KiB (manually set)\n", manual_chunk_size >> 10);
     else {
 #ifdef __linux__
@@ -2411,6 +2528,7 @@ skip_file_scan:
 #else
       fprintf(stderr, "I/O chunk size: %" PRIuMAX " KiB (default size)\n", (uintmax_t)(auto_chunk_size >> 10));
 #endif /* __linux__ */
+#endif /* NO_CHUNKSIZE */
     }
 #ifdef ON_WINDOWS
  #ifndef NO_HARDLINKS
